#!/usr/bin/env python

import re, json, decimal

def check_rx(rx, matches, non_matches):
    for s in matches:
        match = rx.match(s)
        if not match:
            raise Exception('Expected to match [{}]'.format(s))
    for s in non_matches:
        match = rx.match(s)
        if match:
            raise Exception('Did not expect to match [{}] ({})'.format(s, match))

def test_identifiers():
    rx = re.compile(r'[^\W\d\s][\w]*')
    matches = (
        '_charlie',
        '_8eyes',
        'eyes8_',
        'eyes_8'
    )
    non_matches = (
        ' _charlie',
        '8eyes',
        ' spaces',
        '?whatabout',
        '? hmm',
        '2 sup'
    )
    check_rx(rx, matches, non_matches)

def test_type_literals():
    rx = re.compile(r'<[*&@$]?[^\W\d\s][\w]*!?>')
    matches = (
        '<Person>',
        '<Person_14x>',
        '<Person!>',
        '<*Person>',
        '<*Person!>',
        '<&Person>',
        '<&Person!>',
        '<@Person>',
        '<@Person!>',
        '<$Person>',
        '<$Person!>',
        '<Person>>',
    )
    non_matches = (
        '<8Person>',
        '<<Person_14x>',
        '<>Person>',
        '<Person<>',
        '<!Person>',
    )
    check_rx(rx, matches, non_matches)

def test_decimal_literals():
    regexes = (
        re.compile(r'((-?[0-9]+\.[0-9]+)[eE]([-+]?[0-9]+))(f32|f64)?'),
        re.compile(r'((-?[0-9]+\.[0-9]+)[eE]([-+]?[0-9]+))(f32|f64)?'),
        re.compile(r'((-?\.[0-9]+)[eE]([-+]?[0-9]+))(f32|f64)?'),
        re.compile(r'((-?[0-9]+\.)[eE]([-+]?[0-9]+))(f32|f64)?'),
        re.compile(r'((-?[0-9]+)[eE]([-+]?[0-9]+))(f32|f64)?'),
        re.compile(r'(-?[0-9]+\.[0-9])(f32|f64)?'),
        re.compile(r'(-?\.[0-9]+)(f32|f64)?'),
        re.compile(r'(-?[0-9]+\.)(f32|f64)?'),
        re.compile(r'(-?[0-9]+)(f32|f64|d)?'),
    )

    literals = (
        '123.456e-7', '123.456E-7', '123.456e7', '123.456E7', '123.e-7',
        '123.E-7', '123.e7', '123.E7', '123e-7', '123E-7', '123e7', '123E7',
        '.456e-7', '.456E-7', '.456e7', '.456E7', '123.456', '123.', '.456',
        '123.456e-7', '123.456E-7', '123.456e7', '123.456E7', '123.e-7',
        '123.E-7', '123.e7', '123.E7', '123e-7', '123E-7', '123e7', '123E7',
        '.456e-7f32', '.456E-7f32', '.456e7f32', '.456E7f32', '123.456f32',
        '123.f32', '.456f32', '123.456e-7f64', '123.456E-7f64', '123.456e7f64',
        '123.456E7f64', '123.e-7f64', '123.E-7f64', '123.e7f64', '123.E7f64',
        '123e-7f64', '123E-7f64', '123e7f64', '123E7f64', '.456e-7f64',
        '.456E-7f64', '.456e7f64', '.456E7f64', '-123.456e-7', '-123.456E-7',
        '-123.456e7', '-123.456E7', '-123.e-7', '-123.E-7', '-123.e7',
        '-123.E7', '-123e-7', '-123E-7', '-123e7', '-123E7', '-.456e-7',
        '-.456E-7', '-.456e7', '-.456E7', '-123.456', '-123.', '-.456',
        '-123.456e-7', '-123.456E-7', '-123.456e7', '-123.456E7', '-123.e-7',
        '-123.E-7', '-123.e7', '-123.E7', '-123e-7', '-123E-7', '-123e7',
        '-123E7', '-.456e-7f32', '-.456E-7f32', '-.456e7f32', '-.456E7f32',
        '-123.456f32', '-123.f32', '-.456f32', '-123.456e-7f64',
        '-123.456E-7f64', '-123.456e7f64', '-123.456E7f64', '-123.e-7f64',
        '-123.E-7f64', '-123.e7f64', '-123.E7f64', '-123e-7f64', '-123E-7f64',
        '-123e7f64', '-123E7f64', '-.456e-7f64', '-.456E-7f64', '-.456e7f64',
        '-.456E7f64', r'1d'
    )

    d = {}.fromkeys(literals)
    for literal in d:
        for rx in regexes:
            match = rx.match(literal)
            d[literal] = match
            if match:
                found_match = True
                if match.group().endswith('f32'):
                    lit = literal[:-3]
                    print(literal, decimal.Decimal(lit), '32-bit')
                if match.group().endswith('f64'):
                    lit = literal[:-3]
                    print(literal, decimal.Decimal(lit), '64-bit')
                if match.group().endswith('d'):
                    lit = literal[:-1]
                    print(literal, decimal.Decimal(lit), 'Decimal')
                break

def print_regexes():
    hex_prefix = '0[xX]'
    hex_digits = '[0-9a-fA-F]+'
    bin_prefix = '0[bB]'
    bin_digits = '[01]+'
    simple_escape = r"""([a-zA-Z._~!=&\^\-\\?'"])"""
    integer_escape = r'(\d+)'
    hex_escape = r'(x[0-9a-fA-F]+)'
    bad_escape = r"""([\\][^a-zA-Z._~^!=&\^\-\\?'"x0-7])"""
    escape_sequence = r"""(\\("""+simple_escape+'|'+integer_escape+'|'+hex_escape+'))'
    cconst_char = r"""([^'\\\n]|"""+escape_sequence+')'
    char_const = "'"+cconst_char+"'"
    string_char = r"""([^"\\\n]|"""+escape_sequence+')'
    string_literal = '"'+string_char+'*"'
    exponent_part = r"""([eE][-+]?[0-9]+)"""
    fractional_constant = r"""([0-9]*\.[0-9]+)|([0-9]+\.)"""
    floating_constant = '(((('+fractional_constant+')'+exponent_part+'?)|([0-9]+'+exponent_part+'))[fd]?)'
    binary_exponent_part = r'''([pP][+-]?[0-9]+)'''
    hex_fractional_constant = '((('+hex_digits+r""")?\."""+hex_digits+')|('+hex_digits+r"""\.))"""
    hex_floating_constant = '('+hex_prefix+'('+hex_digits+'|'+hex_fractional_constant+')'+binary_exponent_part+'[fd]?)'
    d = {
        'char_const': char_const,
        'string_literal': string_literal,
        'floating_constant': floating_constant,
        'hex_fractional_constant': hex_fractional_constant,
        'hex_floating_constant': hex_floating_constant
    }
    print(json.dumps(d, indent=4, separators=(',', ': ')))

def main():
    test_decimal_literals()
    # test_identifiers()
    # test_type_literals()
    # print_regexes()
    print('\nAll tests pass!')

main()
